//RTDEX_RX_DEMO

#include "stdafx.h"
#include <winioctl.h>

#define BUFSIZE     10*8192

// VHS Definitions => Change these settings for desired board
#define VHS_BT        BT_VHSADC16
#define VHS_ID        0


#include "vhsadac16_regs.h"
#include "vhsadc16ctrl.h"
#include "vhsadc16ctrl_v4.h"
#include "vhsdac16ctrl.h"
#include "vhsadac16_host_api.h"

int main(int argc, char* argv[])
{
    FILE *rxFile;
    PUCHAR inBuff, tmpBuff;
    VHSADC16CONTEXT vhsContext;
    HANDLE hdevice;
    int iStatus;
//    unsigned char ucdata80Mhz[17] = {19,76,64,0,0,32,0,0,32,0,0,0,0,0,0,14,20000000 };
    unsigned char ucdata105Mhz[17] = {16,84,64,0,0,32,0,0,32,0,0,0,0,0,0,14,200 };

//    int i;

    // Beginning of program
    printf("=====================================================\n");
    printf("   VHS-ADAC RTDEx RX Demo, version: 1.2              \n");
    printf("   (c) Lyrtech Signal Processing 2007                \n");
    printf("=====================================================\n\n");

    // Open VHS-ADAC API driver
    iStatus = Vhsadac16_Open(&vhsContext);
    if (iStatus == -1)
    {
        printf( "No valid handle returned by Vhsadac16_Open routine  !!!\n");    
        return -1;
    }
    hdevice = Vhsadac16_getHandle( VHS_BT, VHS_ID );
    if ( hdevice == INVALID_HANDLE_VALUE )
    { 
        printf( "Board %s [%d] not found! \n", (VHS_BT==BT_VHSADC16)?"VHS-ADC":"VHS-DAC",VHS_ID);
        return -1;
    }

    // Allocate memory for transfer
    inBuff = (PUCHAR) malloc(BUFSIZE);
    if ( inBuff == NULL )
    {
        printf("Error allocating input buffer! Exiting...\n");
        return -1;
    }

    // Init inBuff with 0x0
    memset(inBuff, 0, BUFSIZE);
    tmpBuff = inBuff;

    // Open file for uploaded rx data
    rxFile = fopen("rtdex_rx_data.bin", "wb");
    if( rxFile == NULL )
    {
        printf("ERROR OPENING RX FILE!!!\n");
        return -1;
    }

    // set clock to programmable clock
    if( Vhsadac16_isVhsBoardV4(hdevice) )
    {
        printf("\nUsing ADC V4 ONBOARD FIXED clock...\n");
        VHSADC_SetClockSource(hdevice, ONBOARD);
    }
    else
    {
        printf("\nProgramming clock of ADC to 105 MHz\n");
                    
        //enable programmable clock 
        VHSADC_SetProgrammableCLKHardware(hdevice, OUTPUT_ENABLE_PROGRAMMABLE_CLK, SEL_CD_PROGRAMMABLE_CLK_OFF);

        // program the clock frequency
        VHSADC_SetProgrammableCLKFrequency(hdevice, ucdata105Mhz);
    }

    // set ADC run state
    printf("\nSetting ADC Run State\n");
    VHSADC_SetAcquisitionStatus(hdevice,ACQUISITION_ON);

    // set custom reg
    printf("Settings VHS Custom regs (writing 412977624 into cReg0)...\n");
    Vhsadac16_WriteFpgaReg(hdevice, VHSVIRTEX_USER_CTRL0, 412977624);
    
    // Reset and Start RTDEX
    Vhsadac16_RTDExReset(hdevice, FALSE, FALSE);

    // Upload data via RTDEx
    printf("\nUploading data from RTDEx\n");
    
    // Read 10*8192 bytes using RTDEx.
    Vhsadac16_RTDExReadBuffer(hdevice,tmpBuff, BUFSIZE);
        


    {
        int okay;
        VHSADAC_RTDEX_ERROR_INFO errInfo;

        okay = Vhsadac16_GetRTDExError(hdevice, &errInfo);
        if( okay == 0 )
        {
            printf("RTDEx ERRORs REPORT:\n");
            printf("--------------------\n");
            printf("RX OVERRUN  COUNT = %u\n", errInfo.rxOverrun);
            printf("TX UNDERRUN COUNT = %u\n", errInfo.txUnderrun);
            printf("--------------------\n");
        }
        else
        {
            printf("GET_RTDEX_ERROR IOCTL ERROR!!!\n");
        }
    }

    // Write uploaded data to file
    fwrite(inBuff, sizeof(char), BUFSIZE, rxFile);

    // Close file
    fclose(rxFile);

    // Free memory
    free(inBuff);

    // Close driver
    CloseHandle(hdevice);

    printf("\n\nProgram terminated ...press Enter to exit\n");
    while(_kbhit());
    getchar();
    return 0;
}
