#    Control module for the SmartPet Analogue Buffer Card (ABC)

set frame  ABC
set title "SmartPet Analogue Buffer Card"
set icon   ABC

set started 0

set actonall 0

for {set i 0} {$i < 32} {incr i} {set Relay($i) 0; set Bias($i) 0; set Gain($i) 0; set Offset($i) 0;}

set advancedmenu [list \
   "Save Settings to file"  \
   "Restore Settings from file"  \
   "Send Settings to Server"  \
   "Fetch Settings from Server"  \
   "Send Settings to module"  \
   "Load Settings into DACS"  \
   "Update all Settings from module"  \
   "Show module internal memory" \
]


set WebServer ""
set savefile ABCParameters



proc bgerror args {
    global errorInfo tkerrorInfo tkskiperror
    upvar #0 MIDAS_SOAPFaultCode FaultCode
    upvar #0 MIDAS_SOAPFaultString FaultString
    upvar #0 MIDAS_SOAPErrorCode ErrorCode
    upvar #0 MIDAS_SOAPErrorInfo ErrorInfo

    if {$tkskiperror == 1} then {set tkskiperror 0; return}
    set tkerrorInfo $errorInfo

    if {[grab current .] != ""} then {grab release [grab current .]}

    catch {:modify [self frame-name] :busy T}
    if {![info exists FaultCode] || $FaultCode == "NONE"} {
        set z [catch {midas_dialog [self frame-name].tkerrorDialog "Error Detected in Tcl Script" \
                "Task '[Get_My_Name]' detected error in Tcl script\n\n$[lindex $args 0]" "error" 0 "continue" "Stack Trace"} button]
    } else {
        set z [catch {midas_dialog [self frame-name].tkerrorDialog "Error Detected in Tcl Script" \
                "Task '[Get_My_Name]' detected error in Tcl script\n\n$FaultCode\n$FaultString\n$ErrorCode\n$ErrorInfo" \
                       "error" 0 "continue"} button]
    }
    catch {:modify [self frame-name] :busy F}

    if {$z != 0} then {
       puts stderr "midas_dialog: [self frame-name].tkerrorDialog: $button"
       destroy [self frame-name].tkerrorDialog
       return
    }

    if {$button != 0} then {tkStackTrace}
}


proc FetchSettings {args} {                    ;# update parameters from the server
    global Relay Bias Gain Offset

   set-footer "Fetching module settings..."

   set rc [catch {ABCAccessClient__Read Relay} m]
   if {$rc != 0} {bgerror; return}
   for {set i 0} {$i < 32} {incr i} {set Relay($i) [lindex $m $i]}

   set rc [catch {ABCAccessClient__Read Bias} m]
   if {$rc != 0} {bgerror; return}
   for {set i 0} {$i < 32} {incr i} {set Bias($i) [lindex $m $i]}

   set rc [catch {ABCAccessClient__Read Offset} m]
   if {$rc != 0} {bgerror; return}
   for {set i 0} {$i < 32} {incr i} {set Offset($i) [lindex $m $i]}


   set rc [catch {ABCAccessClient__Read Gain} m]
   if {$rc != 0} {bgerror; return}
   for {set i 0} {$i < 32} {incr i} {set Gain($i) [lindex $m $i]}

   Redisplay

   set-footer "Fetched module settings"
}

proc SendSettings {args} {                    ;# load parameters to the server
   global Relay Bias Gain Offset

   set-footer "Sending module settings..."

   for {set i 0} {$i < 32} {incr i} {
        set rc [catch {ABCAccessClient__Set Relay $i $Relay($i)} m]
        if {$rc != 0} {bgerror}
   }

   for {set i 0} {$i < 32} {incr i} {
        set rc [catch {ABCAccessClient__Set Bias $i $Bias($i)} m]
        if {$rc != 0} {bgerror}
   }

   for {set i 0} {$i < 32} {incr i} {
        set rc [catch {ABCAccessClient__Set Offset $i $Offset($i)} m]
        if {$rc != 0} {bgerror}
   }

   for {set i 0} {$i < 32} {incr i} {
        set rc [catch {ABCAccessClient__Set Gain $i $Gain($i)} m]
        if {$rc != 0} {bgerror}
   }

   set-footer "Sent module settings"
}


proc click-bias {i vl} {
    global Bias actonall

    set Bias($i) $vl

    if {$actonall == 0} {
        set rc [catch {ABCAccessClient__Set Bias $i $vl} m]
        :modify DACB$i :value $vl
    } else {
        set rc [catch {ABCAccessClient__Write Bias $vl} m]
        for {set j 0} {$j < 32} {incr j} {
            set Bias($j) $vl
            :modify DACB$j :value $vl
        }
    }

    if {$rc != 0} {bgerror}
}

proc click-gain {i vl} {
    global Gain actonall

    set Gain($i) $vl

    if {$actonall == 0} {
        set rc [catch {ABCAccessClient__Set Gain $i $vl} m]
        :modify DACG$i :value $vl
    } else {
        set rc [catch {ABCAccessClient__Write Gain $vl} m]
        for {set j 0} {$j < 32} {incr j} {
            set Gain($j) $vl
            :modify DACG$j :value $vl
        }
    }

    if {$rc != 0} {bgerror}
}

proc click-offset {i vl} {
    global Offset actonall

    set Offset($i) $vl

    if {$actonall == 0} {
        set rc [catch {ABCAccessClient__Set Offset $i $vl} m]
        :modify DACO$i :value $vl
    } else {
        set rc [catch {ABCAccessClient__Write Offset $vl} m]
        for {set j 0} {$j < 32} {incr j} {
            set Offset($j) $vl
            :modify DACO$j :value $vl
        }
    }

    if {$rc != 0} {bgerror}
}

proc click-relay {i} {
    global Relay actonall

    set Relay($i) [expr ($Relay($i) + 1) & 1]
    set vl $Relay($i)

    if {$actonall == 0} {
        set rc [catch {ABCAccessClient__Set Relay $i $vl} m]
        [:path RELAY].bit$i delete 0 end
        if {$vl == 0} {set c "x2"} {set c "x1"}
        [:path RELAY].bit$i insert 0 "$c"
    } else {
        set rc [catch {ABCAccessClient__Write Relay $vl} m]
        for {set j 0} {$j < 32} {incr j} {
            set Relay($j) $vl
            [:path RELAY].bit$j delete 0 end
            if {$vl == 0} {set c "x2"} {set c "x1"}
            [:path RELAY].bit$j insert 0 "$c"
        }
    }

    if {$rc != 0} {bgerror}
}

proc Redisplay {args} {
    global Relay Bias Gain Offset


   for {set i 0} {$i < 32} {incr i} {
       [:path RELAY].bit$i delete 0 end
       if {$Relay($i) == 0} {set c "x2"} {set c "x1"}
       [:path RELAY].bit$i insert 0 "$c"
   }

   for {set i 0} {$i < 32} {incr i} {
       :modify DACB$i :value $Bias($i)
   }

   for {set i 0} {$i < 32} {incr i} {
       :modify DACG$i :value $Gain($i)
   }

   for {set i 0} {$i < 32} {incr i} {
       :modify DACO$i :value $Offset($i)
   }

}

proc click_save {args} {
   global env savefile
    global Relay Bias Gain Offset

    click_redisplay     ;#    read parameter settings

   set z [catch {open [file join $env(EXPERIMENTBASE) $savefile] w} f]
   if {$z != 0} then {
        midas-warning "Cannot create file [file join $env(EXPERIMENTBASE) $savefile]\n $f"
        return
   }

   for {set i 0} {$i < 32} {incr i} {
       puts $f "Relay$i $Relay($i)"
   }

   for {set i 0} {$i < 32} {incr i} {
       puts $f "Bias$i $Bias($i)"
   }

   for {set i 0} {$i < 32} {incr i} {
       puts $f "Gain$i $Gain($i)"
   }

   for {set i 0} {$i < 32} {incr i} {
       puts $f "Offset$i $Offset($i)"
   }

   close $f

   set-footer "Module settings saved to [file join $env(EXPERIMENTBASE) $savefile]"
}


proc click_restore {args} {
   global env savefile
    global Relay Bias Gain Offset

   set z [catch {open [file join $env(EXPERIMENTBASE) $savefile] r} f]
   if {$z != 0} then {
        midas-warning "Cannot access file [file join $env(EXPERIMENTBASE) $savefile]\n $f"
        return
   }

#    read module settings from file
   set Settings [read $f]
   close $f

   foreach item [split $Settings \n] {
       if {$item == ""} {continue}
       switch -glob [first $item] {
           Relay*  {set Relay([string range [first $item] 5 end]) [second $item]}
           Bias*   {set Bias([string range [first $item] 4 end]) [second $item]}
           Gain*   {set Gain([string range [first $item] 4 end]) [second $item]}
           Offset* {set Offset([string range [first $item] 6 end]) [second $item]}
       }
   }

   set-footer "Module settings restored from [file join $env(EXPERIMENTBASE) $savefile]"

   Redisplay

}

proc click_redisplay {args} {Redisplay}

proc click_applyandload {args} {click_apply; click_load}

proc click_apply {args} {
    set rc [catch {ABCAccessClient__Apply} m]
    if {$rc != 0} {bgerror}
}

proc click_load {args} {
    set rc [catch {ABCAccessClient__Load} m]
    if {$rc != 0} {bgerror}
}

proc click_update {args} {
    global Relay Bias Gain Offset

   set-footer "Fetching module settings..."

    set rc [catch {ABCAccessClient__Status} m]
    if {$rc != 0} {bgerror; return}

    set index 2
    for {set i 0} {$i < 32} {incr i} {

        set lsb [lindex $m $index]
        incr index
        set msb [lindex $m $index]
        incr index
        set Bias($i) [expr ($msb << 8) + $lsb]

        set lsb [lindex $m $index]
        incr index
        set msb [lindex $m $index]
        incr index
        set Gain($i) [expr ($msb << 8) + $lsb]

        set lsb [lindex $m $index]
        incr index
        set msb [lindex $m $index]
        incr index
        set Offset($i) [expr ($msb << 8) + $lsb]
    }

    set v [lindex $m $index]
    incr index

    for {set i 0} {$i < 8} {incr i} {
        set Relay($i) [expr ($v >> $i) && 1]
    }

    set v [lindex $m $index]
    incr index

    for {set i 0} {$i < 8} {incr i} {
        set Relay([expr 8 + $i]) [expr ($v >> $i) & 1]
    }

    set v [lindex $m $index]
    incr index

    for {set i 0} {$i < 8} {incr i} {
        set Relay([expr 16 + $i]) [expr ($v >> $i) & 1]
    }

    set v [lindex $m $index]
    incr index

    for {set i 0} {$i < 8} {incr i} {
        set Relay([expr 24 + $i]) [expr ($v >> $i) & 1]
    }

    SendSettings
    FetchSettings

}

proc click_fetch {args} {FetchSettings}
proc click_send  {args} {SendSettings}

proc click_show {args} {

    set rc [catch {ABCAccessClient__Status} m]
    if {$rc != 0} {bgerror; return}

    midas-information "\
        Control Register (CR) = [lindex $m 0]\n\
        Status Register (SR) = [lindex $m 1]\n\
        Module Internal Memory =\n\
          [lrange $m 2 31]\n\
          [lrange $m 32 61]\n\
          [lrange $m 62 91]\n\
          [lrange $m 92 121]\n\
          [lrange $m 122 151]\n\
          [lrange $m 152 181]\n\
          [lrange $m 182 193]\n\
          [lrange $m 194 197]\n\
    "

    :modify G :cols 132
}


proc click arg {
    global actonall

   set f [lindex $arg 0]
   set c [lindex $arg 1]
   set w [lindex $arg 2]
   set v [lindex $arg 3]
   set i [lindex $arg 4]

   if {$c==":quit"} then {click-quit; return}
   clear-footer
   switch -glob $w {

      ACTONALL   {set actonall $v}

      DACB*      {click-bias [string range $w 4 end] $v}
      DACG*      {click-gain [string range $w 4 end] $v}
      DACO*      {click-offset [string range $w 4 end] $v}

      ACTIONS    {
          switch $i {
            0   {click_save}
            1   {click_restore}
            2   {click_send}
            3   {click_fetch}
            4   {click_apply}
            5   {click_load}
            6   {click_update}
            7   {click_show}
          }
      }

      default           {
           set p "click_[lowercase $w]"
           if {[info procs $p] == "$p"} then {$p [list $v $i]} else {insert-debug "$arg"}
      }
   }
}

proc start args {
   global started Options
   global WebServer ABCAccessServerHost

   if {$started == 0} then {make-frame} else {open-frame}
   set started 1

   load_Options
   if {![info exists Options(WebServer)]} {set Options(WebServer) "localhost"}
   update_Options

   set ABCAccessServerHost $Options(WebServer)
   loadsource ABCAccessClient     ;# better solution needed
   package require ABCAccessClient

   ABCAccessClient__Init
}



proc make-frame {} {
   global env
   global actonall
   global advancedmenu


  :frame [self frame-name] :label "[self frame-title]" \
     :icon-name "[self icon-title]"

  :panel P1 :layout T :padx 5 :pady 5
    :layout H; :col-gap 10; :row-gap 10

        :message M1 :label "Relays" :bold T

        :canvas RELAY  :w 1120 :h 60
        [:path RELAY] config -background $env(tkframebackground)

        set x 5

           for {set i 0} {$i < 32} {incr i} {
              if {[expr ($i/4)*4] == $i} {
                  incr x 10
                  message [:path RELAY].t$i
                  [:path RELAY].t$i configure -text "[expr 1+$i]"
                  :draw window :value $x 40 -window [:path RELAY].t$i
                  incr x 10
              }
              entry [:path RELAY].bit$i -relief flat -width 2 -justify right
              [:path RELAY].bit$i insert 0 "x2"
              :draw window :tag RELAYbit$i :value $x 20 -window [:path RELAY].bit$i 
              bind [:path RELAY].bit$i <ButtonPress-1>  "click-relay $i"
              incr x 30
           }

  :panel P2 :layout L :padx 5 :pady 5
    :layout H; :col-gap 8; :row-gap 8

       :message M1 :label "Input Bias (mV)" :w 28
       :message M2 :label "Gain" :w 20
       :message M3 :label "Output Offset (mV)" :w 32

       :next-row T

       for {set i 0} {$i < 16} {incr i} {
           :message   DACT$i :label "Ch[expr 1+$i]" :w 5 :bold T
           :slider DACB$i :label "" :min +2900 :max -2900 :scale-length 70 :entry-width 4
           :slider DACG$i :label "" :min 1 :max 100 :scale-length 70 :entry-width 3
           :slider DACO$i :label "" :min +1250 :max -1250 :scale-length 70 :entry-width 4

          :next-row T
       }

  :panel P3 :layout R :padx 5 :pady 5
    :layout H; :col-gap 8; :row-gap 8

       :message M1 :label "Input Bias (mV)" :w 28
       :message M2 :label "Gain" :w 20
       :message M3 :label "Output Offset (mV)" :w 32

       :next-row T

       for {set i 16} {$i < 32} {incr i} {
           :message   DACT$i :label "Ch[expr 1+$i]" :w 5 :bold T
           :slider DACB$i :label "" :min +2900 :max -2900 :scale-length 70 :entry-width 4
           :slider DACG$i :label "" :min 1 :max 100 :scale-length 70 :entry-width 3
           :slider DACO$i :label "" :min +1250 :max -1250 :scale-length 70 :entry-width 4

          :next-row T
       }

  :panel BUTTONS :after footer :layout B :padx 5 :pady 5
    :layout H; :col-gap 20; :row-gap 10

     :button REDISPLAY
     :button APPLYANDLOAD   :label "Apply Settings"

     :menubutton ACTIONS :label "Advanced"  :strings $advancedmenu

       :hskip 50
       :checkbox ACTONALL :label "Act on all channels?" :bold T :value $actonall

   :show T
}

#    The Options are experiment specific and normally used by SYS
#    However the same code as for the session preferences is used which allows for wider use in the future.

proc get_Option_file {} {
    global env

#    look for the Option file
    set pd $env(EXPERIMENTBASE)
    if {![file exists $pd]} then {
         insert-log "$pd does not exist - Join an experiment!"
         return {}
    }
    if {![file isdirectory $pd]} then {
         insert-log "$pwd is not useable - it is not a directory"
         return {}
    }
    set pf [file join $pd MidasOptions]
    set z [catch {set f [open $pf r]} m]
    if {$z != 0} then {
        set z [catch {set f [open $pf w]} m]
        if {$z != 0} then {
            insert-log "Unable to access $pf - $m"
            return 1
        }
        close $f
        return {}
    }

#    and read it
    set lines [split [read $f] "\n"]
    close $f

    return $lines
}

proc put_Option_file {lines} {
   global env

#    look for the Option file
    set pd $env(EXPERIMENTBASE)
    if {![file exists $pd]} then {
         insert-log "$pd does not exist - Join an experiment!"
         return {}
    }
    if {![file isdirectory $pd]} then {
         insert-log "$pwd is not useable - it is not a directory"
         return {}
    }
    set pf [file join $pd MidasOptions]
    set z [catch {set f [open $pf w]} m]
    if {$z != 0} then {
        insert-log "Unable to access $pf - $m"
        return 1
    }

    foreach item $lines {
      if {$item != ""} {puts $f $item}
    }

    close $f
    return 0
}

#    save all Options for the current task
proc update_Options {} {
   global env Options

   set lines [get_Option_file]
   set key [Get_My_Name]
   set range [find-keyword $lines $key]
   if {[llength $range] > 0} then {
       set lines [lreplace $lines [first $range] [expr [second $range] + 1]]
   }

   lappend lines "## $key"
   foreach item [array names Options] {
       lappend lines "$item $Options($item)"
   }
   lappend lines "##"

   set z [put_Option_file $lines]
   return $z
}

#    load all Options for the current task
proc load_Options {} {
   global env Options

   set lines [get_Option_file]
   set key [Get_My_Name]
   set range [find-keyword $lines $key]

   if {[llength $range] > 1} then {
       set stuff [lrange $lines [expr [first $range]+1] [second $range]]
       foreach item $stuff {
           set Options([head $item]) [tail $item]
       }
   }
}

#    return task Options
proc get_Options {task} {
   global env

   set lines [get_Option_file]
   set key $task
   set range [find-keyword $lines $key]

   set Options {}

   if {[llength $range] > 1} then {
       set stuff [lrange $lines [expr [first $range]+1] [second $range]]
       foreach item $stuff {
           lappend Options [list [first $item] [second $item]]
       }
   } else {
       lappend Options [list Options.filename  "[file join $env(HOME) MidasOptions Options]"]
   }
   return $Options
}


