#    Diagnostic and test task for the  Lyrtech VHS-ADC 

set frame  VHSADC
set title "Lyrtech VHS-ADC Client"
set icon   VHSADC

set started 0

set inputcolour  green4
set outputcolour red
set insertcolour blue
set bkgcolour    azure

set LogFlag 1

set verbose 0

set VHSADCServerList {}
set VHSADCServer "select a server"
variable WebServer

set AcquisitionMethod 0
set ADCClockFreq 105
set FrameSize 512
set FPGAProgramOption 1

set boardID -1
set maxboards -1
set boardlist "none"

set numchans 16
set chanlist {}

for {set i 0} {$i < 8} {incr i} {set NumChan($i) $numchans}

set chan  0

set allchans 0
set allmodules 0

set waveform "undefined"
set directory $env(HOME)

set fpgadirectory ""
set fpgaffile ""

set clkfile ""
set clkdirectory ""

set savefile ""
set savedirectory ""

proc VHSADCClient__Vhsadac16ReadFpgaReg  {args} {
    midas-warning "Please select a VHSADC WebServer"
    error "No VHSADC WebServer selected"
}

proc VHSADCClient__Vhsadac16WriteFpgaReg  {args} {
    midas-warning "Please select a VHSADC WebServer"
    error "No VHSADC WebServer selected"
}

proc VHSADCClient__Vhsadac16Sdram_Read  {args} {
    midas-warning "Please select a VHSADC WebServer"
    error "No VHSADC WebServer selected"
}

proc bgerror err {
    global errorInfo tkerrorInfo tkskiperror
    upvar #0 MIDAS_SOAPFaultCode FaultCode
    upvar #0 MIDAS_SOAPFaultString FaultString
    upvar #0 MIDAS_SOAPErrorCode ErrorCode
    upvar #0 MIDAS_SOAPErrorInfo ErrorInfo

    if {$tkskiperror == 1} then {set tkskiperror 0; return}
    set tkerrorInfo $errorInfo

    if {[grab current .] != ""} then {grab release [grab current .]}

    catch {:modify [self frame-name] :busy T}
    if {![info exists FaultCode] || $FaultCode == "NONE"} {
        set z [catch {midas_dialog [self frame-name].tkerrorDialog "Error Detected in Tcl Script" \
                "Task '[Get_My_Name]' detected error in Tcl script\n\n$err" "error" 0 "continue" "Stack Trace"} button]
    } else {
        set z [catch {midas_dialog [self frame-name].tkerrorDialog "Error Detected in Tcl Script" \
                "Task '[Get_My_Name]' detected error in Tcl script\n\n$FaultCode\n$FaultString\n$ErrorCode\n$ErrorInfo" \
                       "error" 0 "continue"} button]
    }
    catch {:modify [self frame-name] :busy F}

    if {$z != 0} then {
       puts stderr "midas_dialog: [self frame-name].tkerrorDialog: $button"
       destroy [self frame-name].tkerrorDialog
       return
    }

    if {$button != 0} then {tkStackTrace}
}

proc make-frame {} {
   global env

set rows 12
set cols 64

  :frame [self frame-name] :label "[self frame-title]" \
     :icon-name "[self icon-title]"

  :panel P0 :layout V :padx 5 :pady 5
    :layout H; :col-gap 10; :row-gap 10

    :text WEBSERVERT         :label "VHSADC WebServer" :w 20  :value "select a server"
    :choice-stack WEBSERVERC :label "" :strings ""
    :hskip 80
    :checkbox VERBOSE        :label "Enable Verbose Option" 

    :next-row T
    :number MODULE-NUM :label "VHS-ADC Board Selection:" :min -1 :max -1 :w 5 :value -1
    :menubutton MODULE-SEL :label "" :menu "none"
    :button INFO    :label "Report Board Information"  :command click-boardinfo
    :button OPTIONS :label "Reload Options" :command click-options

  :panel P1 :layout V :padx 5 :pady 5
    :galley LOG :rows $rows :cols $cols :font fixed
     configuretags

  :panel P2 :layout V :padx 5 :pady 5
    :layout H; :col-gap 20; :row-gap 10

     :hskip 40
      :checkbox TRIG :label "Digital Trigger Enabled" :deaf T 
      :checkbox ACQ  :label "Acquisition RUN State" :deaf T
      :button UPDATE :label "Update" :command update_

  :panel P3 :layout V :padx 20 :pady 5
    :layout H; :col-gap 20; :row-gap 10

    :button CLEAR   :label "Clear log window"     :command ":modify LOG :empty T"
    :button REPORT  :label "Show board registers" :command click-boardregisters
    :button SAVE    :label "Save Settings"        :command click-save
    :button RESTORE :label "Restore Settings"     :command click-restore
    :next-row T

    :button P51 :label "FPGA Programming" :w 20 :command do_fpga_programming
    :button P52 :label "Gain Control"     :w 20 :command do_gain_control
    :button P53 :label "User Registers"   :w 20 :command do_user_registers

    :next-row T
    :button P54 :label "Clock settings"   :w 20 :command do_clock
    :button P55 :label "ADC status"       :w 20 :command do_adc
    :button P56 :label "FPGA Acquisition" :w 20 :command do_fpgaacq

    :next-row T
    :button P57 :label "SDRAM"            :w 20 :command do_sdram
    :button P58 :label "Load Default Setup" :w 20 :command do_defaultsetup
    :button P59 :label "Extras"           :w 20 :command do_extras

   open-frame

  :galley_size LOG $rows $cols

   :show T
}

proc do_fpga_programming {} {
    upvar #0 fpgaffile filename
    global ADCClockFreq FrameSize
    global FPGAProgramOption

   set clkchoices [list External 125Mhz 120Mhz 115Mhz 110Mhz 105Mhz 100Mhz 95Mhz 90Mhz 85Mhz 80Mhz 75Mhz 70Mhz 65Mhz 60Mhz 55Mhz 50Mhz 45Mhz 40Mhz 35Mhz 30Mhz]
   set fchoices [list 8 16 32 64 128 256 512 1024]
   
  :frame "[self frame-name].fpga" :label "FPGA programming" :resize F :transient T

   :panel P :layout T :padx 5 :pady 5
     :layout H; :row-gap 10; :col-gap 10

      :text   FPGAFILE    :label "Bitstream file for FPGA" :w 60 :value $filename
      :button FPGABROWSE :label "Browse"

      :next-row T
      set f [lsearch $clkchoices [set ADCClockFreq]Mhz]
      if {$f == -1} {set f 0}
      :choice-stack FPGACLK  :label "ADC Clock"  :strings $clkchoices  :value $f
      :button FPGASETCLK  :label "Program ADC Clock Now"
      :choice FPGAOPT :label ""  :strings [list "Program Current Module" "Program All Modules"] :value $FPGAProgramOption :command do_fpgaOption
      :next-row T
      set f [lsearch $fchoices $FrameSize]
      if {$f == -1} {set f 6}
      :choice-stack FPGAFRAMESIZE  :label "Frame Size"  :strings $fchoices  :value $f
      :button FPGACHANS :label "Check FPGA supported channels"  :command do_fpgasetnumchan
      :next-row T
      :col-gap 20
      :button FPGADONE  :label "Done" :command "frame-quit [self frame-name].fpga"
      :button FPGAUPLOAD    :label "Upload FPGA bitstream file"  :command do_UpLoadfpga
      :button FPGAPGM       :label "Program FPGA"                :command do_Programfpga
      :button FPGAUPLOADPGM :label "Upload & Program FPGA"       :command do_UpLoadProgramfpga
      :button FPGARESET :label "FPGA Internal Reset" :command click-fpgareset

  :show T
}

proc do_gain_control {} {

  :frame "[self frame-name].gain" :label "Programmable gain configuration" :resize F :transient T

   :panel P :layout T :padx 5 :pady 5
     :layout H; :row-gap 10; :col-gap 10

     :choice-stack GAINCTRL :label "Programmable Gain Control" :strings [list register internal]
     :next-row T

      for {set i 1} {$i <= 16} {incr i} {
          :number GAINVL$i :label Ch$i :min 0 :value 4095 :valuex 6c :w 4
          :checkbox GAINEN$i :label Enable
          :next-row T
      }

      :checkbox GAINALL :label "All gains as Ch1"
      :next-row T
      :button GAINPGM   :label "Set programmable gains"

      :next-row T
      :button GAINDONE  :label "Done" :command "frame-quit [self frame-name].gain"
      :button GAINUPDTE :label "Update" :command update_gain

  :show T

   update_gain
}

proc do_extras {} {
   global waveform

  :frame "[self frame-name].extra" :label "User defined control registers" :resize F :transient T

   :panel P :layout T :padx 5 :pady 5
     :layout H; :row-gap 10; :col-gap 10

    :text   WAVEFORM  :label "Waveform file" :w 40  :value $waveform
    :button BROWSEWAVEFORM :label "Browse"
    :button SHOWWAVEFORM   :label "Display"

    :next-row T

      :button EXTRASDONE  :label "Done" :command "frame-quit [self frame-name].extra"
      :button EXTARSUPDTE :label "Update" :command update_extras

  :show T

   update_extras
}

proc FpgaType {} {
    global boardID

#    Get CPLD Registers Information

    set rc [VHSADCClient__Vhsadac16GetCpldRegs $boardID]

    set sm_board_id [lindex $rc 5]

    switch $sm_board_id {
         0x82  -
         0x83  -
         0x84  -
         0x85  {set Type 0}
         0x8a  -
         0x8b  -
         0x8c  -
         0x8d  -
         0x8e  {set Type 1}
         0x8f  -
         default  {set Type -1}
    }

    return $Type
}

proc do_clock {} {

    if {[FpgaType] == 0} {do_clock2} else {do_clock4}
}

proc do_clock2 {} {

  :frame "[self frame-name].clock" :label "Clock Settings (Virtex II)" :resize F :transient T

  :panel P :layout V :padx 5 :pady 5
    :layout H; :col-gap 10; :row-gap 10

      :choice-stack CLKSRC :label "Clock Source:" :strings [list "Programmable (FS6377)" External]
      :hskip 25
      :text DCM :label "DCMs" :w 8
      :hskip 25
      :button RESETDCM :label "Reset FPGA DCM" :command ResetDCM

      :next-row T
      :choice-stack CDPCLK :label "CD PCLK"  :strings [list "PLLC/D(1) and MUXC/D(1)" "PLLC/D(2) and MUXC/D(2)"]
      :choice-stack OEPCLK :label "OE PCLK"  :strings [list Disabled Enabled]
      :choice-stack PDPCLK :label "PCLK O/P" :strings [list "Normal" "Power Down"]

      :next-row T
      :choice-stack QUADCLK :label "Quadrature"      :strings [list OFF ON]
      :choice-stack FPDPCLK :label "FPDP Mastering"  :strings [list "Clk Src from CLK SRC" "Clk Src from FPDP"]

      :next-row T
      :message CLKM1 :label "Programmable Clock Value Registers" :bold T
      :next-row T
      for {set i 1} {$i <= 16} {incr i} {
          :number CLKVL$i :label $i :min 0 :value 0xff :valuex 6c :w 6
          if {[expr ($i/4)*4] == $i} {:next-row T}
      }

      :button CLKPGM    :label "Program FS6377 from register values" 
      :hskip 25
      :button CLK105    :label "Set ALL ADC Clocks to 105MHz"

      :next-row T
      :text   CLKFILE   :label "Dat file for FS6377 Programmable Clock" :w 32 :valuex 32c
      :button CLKBROWSE :label "Browse"
      :button CLKLOAD   :label "Load"

    :next-row T

      :button CLKDONE  :label "Done" :command "frame-quit [self frame-name].clock"
      :button CLKUPDTE :label "Update" :command update_clock

  :show T

   update_clock
}


proc do_clock4 {} {

  :frame "[self frame-name].clock" :label "Clock Settings (Virtex 4)" :resize F :transient T

  :panel P :layout V :padx 5 :pady 5
    :layout H; :col-gap 10; :row-gap 10

      :choice-stack CLKSRC :label "Clock Source:" :strings [list "Onboard (104MHz)" External "Onboard (52MHz)" \
              Reserved "FPDP Rx Clock" "RapidCHANNEL Rx Clock" "FPGA generated"  "FPGA generated"]
      :hskip 25
      :text DCM :label "DCMs" :w 8
      :hskip 25
      :button RESETDCM :label "Reset FPGA DCM" :command ResetDCM

    :next-row T
      :checkbox CLKDIV2 :label "DIF Clock Divided By 2" :deaf T

    :next-row T
      :button CLKDONE  :label "Done" :command "frame-quit [self frame-name].clock"
      :button CLKUPDTE :label "Update" :command update_clock

  :show T

   update_clock
}


proc do_adc {} {

  :frame "[self frame-name].adc" :label "ADC Settings" :resize F :transient T

  :panel P :layout V :padx 5 :pady 5
    :layout H; :col-gap 10; :row-gap 10

    :message ADCM1 :label "ADC Out Of Range Status" :bold T
     :next-row T

      for {set i 1} {$i <= 16} {incr i} {
          :checkbox ADC$i :label Ch$i :valuex 6c 
          if {[expr ($i/4)*4] == $i} {:next-row T}
      }

    :next-row T

      :button ADCDONE  :label "Done" :command "frame-quit [self frame-name].adc"
      :button ADCUPDTE :label "Update" :command update_adc

  :show T

   update_adc
}

proc do_fpgaacq {} {
    global AcquisitionMethod

   if {$AcquisitionMethod == 0} {do_SDRAMfpgaacq; return}
   if {$AcquisitionMethod == 1} {do_RTDEXfpgaacq; return}

}

set sdramaddr  0
set sdramwords 128

set sdrambase 0x0000
set sdrambasechoice [list "SDRAM - 0x0000"]

proc do_sdram {} {
   global sdrambase sdramaddr sdramwords sdrambasechoice

  :frame [self frame-name].sdram :label "Display Memory" :resize F :transient T

    :panel SDRAM :layout T :padx 10 :pady 10
     :layout H; :col-gap 10; :row-gap 10
     :choice-stack SDRAMBASEC  :label "Area base address" :strings $sdrambasechoice :value 0 :valuex 24c
     :text SDRAMBASET :label "" :value $sdrambase :w 12
     :next-row T
     :text SDRAMADDR    :label "Address Offset"  :value $sdramaddr   :valuex 24c  :w 12
     :button SDRAMADDRZ :label "Reset to zero"
     :next-row T
     :layout V
     :text SDRAMWORDS   :label "Words to read"   :value $sdramwords  :valuex 24c  :w 12
       :message SDRAMM1 :label "Memory Contents" :bold T
       :galley SDRAMINFO :rows 20 :cols 110 :deaf T :font fixed

  :panel P3 :layout V :padx 5 :pady 5
    :layout H; :col-gap 20

       :button SDRAMDONE :label "Done" :command "frame-quit [self frame-name].sdram"
       :button SDRAMREAD :label "Read Memory"  :command  do_sdramdisplay
       :button SDRAMCLR  :label "Clear Window" :command ":modify SDRAMINFO :append F"
  
   :show T
}

proc do_SDRAMfpgaacq {} {
    global fpgaacq

  :frame "[self frame-name].fpgaacq" :label "SDRAM FPGA Acquisition" :resize F :transient T

  :panel P1 :layout V :padx 5 :pady 5
    :layout H; :col-gap 10; :row-gap 10

      :message FPGAACQP1 :label "Acquisition Options" :bold T
      :next-row T

      :choice-stack FPGAACQSF  :label "Storage Format" :strings [list Single Double Quad]
      :choice-stack FPGAACQSID :label "Storage ID"     :strings [list "Ext SDRAM"]
      :choice-stack FPGAACQSRC :label "Source ID"      :strings [list Internal VHS-ADC FPGALink]

      :next-row T
      :number FPGAACQFRAMESIZE :label "Frame Size" :min 32 :max 8192 :valuex 12c :w 6

  :panel P2 :layout V :padx 5 :pady 5
    :layout H; :col-gap 10; :row-gap 10

      :message FPGAACQP2 :label "Trigger Control" :bold T
      :next-row T

      :choice-stack FPGAACQTRIGSEL :label "Select" \
          :strings [list Ch0 Ch1 Ch2 Ch3 Ch4 Ch5 Ch6 Ch7 Ch8 Ch9 Ch10 Ch11 Ch12 Ch13 Ch14 Ch15] :valuex 8c
      :choice-stack FPGAACQTRIGPOL :label "Polarity" :strings [list Falling Rising] :valuex 8c
      :choice-stack FPGAACQTRIGSRC   :label "Source" :strings [list Manual Signal External] :valuex 8c
      :next-row T
      :choice-stack FPGAACQTRIGMDE   :label "Mode" :strings [list "One Shot   " Continuous] :valuex 8c
      :number FPGAACQTRIGLVL   :label "Level" :min 0 :max 65536 :w 5 :valuex 6c
      :number FPGAACQTRIGDLY   :label "Delay" :min 0 :max 65536 :w 5 :valuex 6c


  :panel P3 :layout V :padx 5 :pady 5
    :layout H; :col-gap 10; :row-gap 10

      :message FPGAACQP3 :label "Channel Select" :bold T
      :next-row T

      for {set i 1} {$i <= 16} {incr i} {
          :checkbox FPGAACQCH$i :label $i :valuex 2c
          if {$i == 8}  {:button FPGAACQALL  :label All  :w 5; :next-row T}
          if {$i == 16} {:button FPGAACQNONE :label None :w 5}
      }


  :panel P4 :layout V :padx 5 :pady 5
    :layout H; :col-gap 20; :row-gap 10

      :message FPGAACQP4 :label "Acquisition Status" :bold T
      :next-row T

      :choice-stack FPGAACQTRIGRUN   :label "Trigger State"  :strings [list Stopped Running]
      :choice-stack FPGAACQTRIGSTATE :label "Acquisition State" \
           :strings [list Idle Armed Waiting Storing Ready Pre-ready "" "" "Data select error" \
                "Trigger select error" "Storage overrun error" "Packing overrun error" "Data overrun error" ]
      :hskip 75
      :button FPGAAUTO :label "Auto Update"  :command fpga_auto_update

      :next-row T
      :text FPGAACQFRAMECNT    :label "Frame Count" :w 6
      :text FPGAACQSAMPLES     :label "Samples" :w 2
      :checkbox FPGAACQFERR :label "Data Framing Error"
      :checkbox FPGAACQACT  :label "Data Available"

  :panel P5 :layout V :padx 5 :pady 5
    :layout H; :col-gap 20

      :button FPGAACQDONE  :label "Done" :command "frame-quit [self frame-name].fpgaacq"
      :button FPGAACQUPDTE :label "Update" :command update_fpgaacq
      :button FPGAACQREGS  :label "Show FPGA Acquisition Registers" :command click-fpgaregisters

  :show T

   update_fpgaacq
}

proc do_RTDEXfpgaacq {} {
    global fpgaacq

  :frame "[self frame-name].fpgaacq" :label "RTDEX FPGA Acquisition" :resize F :transient T

  :panel P1 :layout V :padx 5 :pady 5
    :layout H; :col-gap 10; :row-gap 10

      :message FPGAACQP1 :label "Acquisition Status/Control" :bold T
      :next-row T

      for {set offset 32} {$offset < 42} {incr offset} {
          :text RTDEXACQ$offset :label "register address $offset; [format 0x%x $offset]" :w 6 
          :next-row T
      }

  :panel P5 :layout V :padx 5 :pady 5
    :layout H; :col-gap 20

      :button FPGAACQDONE  :label "Done" :command "frame-quit [self frame-name].fpgaacq"
      :button FPGAACQUPDTE :label "Update" :command update_fpgaacq
      :button FPGAACQREGS  :label "Show FPGA Acquisition Registers" :command click-rtdexregisters

  :show T

   update_fpgaacq
}

proc do_defaultsetup {args} {    ;#  load a setup compatible with our data acquisition program
    global boardlist boardID

    do_SDRAMfpgaacq
    set save $boardID

    foreach boardID $boardlist {

        click-fpgareset
        click-fpgaacq SF 1 Double       ;# data format - 2 samples per word
        click-fpgaacq SRC 1 FPGALink
        click-fpgaacq FRAMESIZE 512 0   ;# Sample length - 512 items
        click-fpgaacq CH1 1 1           ;# Select channel 1
        click-fpgaacq TRIGSRC 2 0       ;# Trigger Source - External 
        click-fpgaacq TRIGMDE 0 0       ;# Trigger Mode - One shot
    }

    set boardID $save     ;# restore 
}

proc do_acqGO {opt} {   ;#  for external control;  opt=0  STOP;   opt!=0  GO
    global boardlist
    global AcquisitionMethod

   if {$AcquisitionMethod == 1} {return}

    if {$opt == 0} then {set vl 0} else {set vl 9}

    foreach boardID $boardlist {
        VHSADCClient__Vhsadac16WriteFpgaReg $boardID 0x1c $vl ;# acq/trig control register
    }
}

#proc do_acqARM {opt} {   ;#  for external control;  opt=0  STOP;   opt!=0  GO
#    global boardlist

#    if {$opt == 0} then {set vl 0} else {set vl 1}

#    foreach boardID $boardlist {
#        set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID [expr 0x1000 + 0x10]]
#        VHSADCClient__Vhsadac16WriteFpgaReg $boardID [expr 0x1000 + 0x10] \
#                [expr ($rc & 0x0018ffff) | ($vl << 17)]   ;# fpga trig control
#   }
#}

proc get_framesize {vl} {   ;#  for external control
    global AcquisitionMethod

   if {$AcquisitionMethod == 0} {return [VHSADCClient__Vhsadac16ReadFpgaReg $vl [expr 0x1000 + 0x08]]}
   if {$AcquisitionMethod == 1} {
        set reg 0x30
        set chan 0
        set rc [VHSADCClient__ReadUserReg $chan $reg]
        set rc [expr $rc & 0xffff]
        if {$rc < 4 || $rc > 1024} {set rc 0x200} else {set rc [expr $rc * 2]}   ;# set a default if value return is invalid
        return [expr $rc - 8]
    }
}

proc get_numchans {vl} {   ;#  for external control
    global NumChan

    return $NumChan($vl)
}

proc redisplay {} {
    catch {update_}
    catch {update_gain}
    catch {update_user}
    catch {update_extras}
    catch {update_clock}
    catch {update_adc}
    catch {update_fpgaacq}
}

proc click-configure {W} {:galley_resize LOG}

proc configuretags {} {
   global inputcolour outputcolour insertcolour bkgcolour

    :modify LOG :createtag "taginput  -foreground $inputcolour  -background $bkgcolour"
    :modify LOG :createtag "tagoutput -foreground $outputcolour -background $bkgcolour"
    :modify LOG :createtag "taginsert -foreground $insertcolour -background $bkgcolour"
}

proc insert {s} {:modify LOG :usingtag taginsert :strings $s\n; update idletasks}


proc start args {
   global started env Options VHSADCServerList
   global numchans chanlist

   if {$started == 0} then {make-frame} else {open-frame}
   set started 1

#    load application options and establish default values for any not defined

   load_Options

   if {![info exists Options(fpgadirectory)]} {set Options(fpgadirectory) $env(HOME)}
   if {![info exists Options(fpgaffile)]}     {set Options(fpgaffile) "undefined"}
   if {![info exists Options(clkdirectory)]}  {set Options(clkdirectory) $env(HOME)}
   if {![info exists Options(clkffile)]}      {set Options(clkfile) "undefined"}
   if {![info exists Options(savedirectory)]} {set Options(savedirectory) $env(HOME)}
   if {![info exists Options(savefile)]}      {set Options(savefile) "undefined"}
   if {![info exists Options(WebServer)]}     {set Options(WebServer) "localhost"}
   if {![info exists Options(AutoSelect)]}     {set Options(AutoSelect) "1"}

   if {![info exists Options(UserGUIVersion)]} {set Options(UserGUIVersion) "0"}
   if {![info exists Options(FpgaRegisters)]}  {set Options(FpgaRegisters) [file join $env(MIDASBASE) LyrTech tcl FpgaRegisters.cfg]}

   if {![info exists Options(AcquisitionMethod)]} {set Options(AcquisitionMethod) "0"}     ;#  0 = SDRAM; 1 = RTDEx

   if {![info exists Options(SaveClocks)]} {set Options(SaveClocks) "0"}
   if {![info exists Options(SaveGains)]} {set Options(SaveGains) "0"}
   if {![info exists Options(SaveUserRegs)]} {set Options(SaveUserRegs) "0"}
   if {![info exists Options(SaveAcqRegs)]} {set Options(SaveAcqRegs) "1"}
   if {![info exists Options(ADCClockFreq)]} {set Options(ADCClockFreq) "105"}
   if {![info exists Options(FrameSize)]} {set Options(FrameSize) "512"}

   update_Options

   foreach option [list WebServer AcquisitionMethod UserGUIVersion fpgadirectory fpgaffile clkdirectory clkfile savedirectory savefile] {
       global $option
       set $option $Options($option)
   }

   set chanlist {}
   for {set i 0} {$i < $numchans} {incr i} {lappend chanlist $i}

   set VHSADCServerList [list $WebServer]    ;# put default into the menu
   :modify WEBSERVERC :strings $VHSADCServerList :value [lsearch $VHSADCServerList $WebServer]

   if {$AcquisitionMethod == 1} {:modify P57 :inactive T; :modify P54 :inactive T}

   if {$Options(AutoSelect) == 1} {click-server $WebServer}

}

proc Save_Options {} {
   global Options

   foreach option [list WebServer AcquisitionMethod UserGUIVersion ADCClockFreq FrameSize fpgadirectory fpgaffile clkdirectory clkfile savedirectory savefile] {
       global $option
       set Options($option) [set $option]
   }

   update_Options
}

proc click-options {} {
   global Options

   load_Options

   foreach option [list WebServer AcquisitionMethod UserGUIVersion ADCClockFreq FrameSize fpgadirectory fpgaffile clkdirectory clkfile savedirectory savefile] {
       global $option
       set $option $Options($option)
   }
}

proc click arg {
   global verbose WebServer
   global allmodules allchans

   set w [lindex $arg 2]
   set v [lindex $arg 3]
   set i [lindex $arg 4]

   clear-footer

   switch -glob $w  {

      WEBSERVERT    {click-server $v}
      WEBSERVERC    {click-server $i}

      MODULE-SEL    -
      MODULE-NUM    {click-select-module $v}

      CHAN-SEL    -
      CHAN-NUM      {click-select-chan $v; return}

      ALL-MODULES   {set allmodules $v; return}
      ALL-CHANS     {set allchans $v; return}

      GAINVL*       {click-gainvl [string range $w 6 end] $v}
      GAINEN*       {click-gainen [string range $w 6 end] $v}

      USERVL*       {click-uservl [string range $w 6 end] $v}
      USER2*        {click_user_2 [string range $w 5 end] $v $i; return}
      MWD*          {click_user_2 [string range $w 3 end] $v $i; return}
      USER*         {click_user_1 [string range $w 4 end] $v; update_user_1; return}

      CLKVL*        {click-clkvl [string range $w 5 end] $v}

      FPGAACQ*      {click-fpgaacq [string range $w 7 end] $v $i}
      RTDEXACQ*     {click-rtdexacq [string range $w 8 end] $v $i}

      FPGACLK       {click-fpgaclk $v $i}
      FPGASETCLK    {click-fpgasetclk}

      FPGAFRAMESIZE {click-fpgaframesize $i}

      VERBOSE       {set verbose $v; if {$WebServer != ""} then {VHSADCClient__VHSADCVerbose $verbose}}
 
      default {
         set p "click-[lowercase $w]"
         if {[info procs $p] == "$p"} then {$p $v} else {set-footer "no handler for widget $w"}
      }
   }

    redisplay 
}

proc click-server {v} {
    global env VHSADCServerList WebServer VHSADCHost
    global boardID maxboards boardlist
    global verbose started

    global VHSADCHost VHSADCPort

    set WebServer $v
    set VHSADCServerList [list $WebServer]    ;# put default into the menu
    :modify WEBSERVERT :value $WebServer
    :modify WEBSERVERC :strings $VHSADCServerList :value [lsearch $VHSADCServerList $WebServer]

    Save_Options

    set VHSADCHost $WebServer
    loadsource VHSADCClient     ;# better solution needed

    VHSADCClient__VHSADCVerbose $verbose
    set maxboards [VHSADCClient__VHSADCInit]
    
    set boardlist {}
    for {set i 0} {$i < $maxboards} {incr i} {lappend boardlist $i}
    set boardID [lindex $boardlist 0]
    :modify MODULE-SEL :menu $boardlist
    :modify MODULE-NUM :min 0 :max [expr $maxboards-1]
    :modify MODULE-NUM :value $boardID

    insert "Initialised VHSADC Web Service at $WebServer: $maxboards VHS-ADC boards discovered"

    set started 2
}

proc click-select-module {vl} {
    global boardID maxboards

    set boardID $vl
    if {$boardID < 0} {set boardID 0}
    if {$boardID >= $maxboards} {set boardID [expr $maxboards-1]}

    :modify MODULE-NUM :value $boardID
}

proc click-save {} {
    global env boardlist chanlist Options NumChan
    global AcquisitionMethod UserGUIVersion
    global boardID chan
    upvar #0 savedirectory  directory
    upvar #0 savefile filename


    set state [inform sys inquire readoutstate]
    if {$state != 0} {
        midas-warning "You can only do this reliably when STOPped"
        return
    }

    	option add *Canvas.background  #c0c020
	set types {
	{ {Data Files} {*.dat} }
	{ {All Files } {*} }
	}

	set f [tk_getSaveFile -parent [self frame-name] \
		  -title "Save File Selection"  \
		  -initialdir $directory -filetypes $types]
	option add *Canvas.background  $env(tkcanvasbackground)
	if {$f == ""} then {return}

      if {[file extension $f] == ""} then {append f .dat}

      set filename $f   ;#  remember current choice
      set directory [file dirname $f]
      Save_Options

      set z [catch {open $filename w} f]
      if {$z != 0} then {
          midas-warning "Cannot create/open file $filename\n$f"
          return
      }

      set fpgatype [FpgaType]

      if {$fpgatype == 0 && $Options(SaveClocks) == 1} {       ;# for Virtex II

          foreach ID $boardlist {

             foreach offset [list 0x8 0xc 0x10 0x14 0x18] {

                  set z [catch {VHSADCClient__Vhsadac16ReadFpgaReg $ID $offset} rc]
                  if {$z == 0} {
                      puts $f "$ID [format 0x%03x $offset] [format 0x%08x $rc]"
                  } else {
                      set-footer "Error reading FpgaReg offset=[format 0x%03x $offset]: NOT saved"
                  }
              }
          }
      }

      if {$fpgatype == 1 && $Options(SaveClocks) == 1} {       ;# for Virtex 4

          foreach ID $boardlist {

                  set offset 0x8
                  set z [catch {VHSADCClient__Vhsadac16ReadFpgaReg $ID $offset} rc]
                  if {$z == 0} {
                      puts $f "$ID [format 0x%03x $offset] [format 0x%08x [expr ($rc & 7) + 0x2000]]"
                  } else {
                      set-footer "Error reading FpgaReg offset=[format 0x%03x $offset]: NOT saved"
                  }
          }
      }

      if {$Options(SaveGains) == 1} {

          foreach ID $boardlist {

             foreach offset [list 0x24 0x28 0x2c 0x30 0x34 0x38 0x3c 0x40 0x44] {

                  set z [catch {VHSADCClient__Vhsadac16ReadFpgaReg $ID $offset} rc]
                  if {$z == 0} {
                      puts $f "$ID [format 0x%03x $offset] [format 0x%08x $rc]"
                  } else {
                      set-footer "Error reading FpgaReg offset=[format 0x%03x $offset]: NOT saved"
                  }
              }
          }
      }

      if {$Options(SaveUserRegs) == 1} {

          foreach ID $boardlist {

             foreach offset [list 0x70 0x74 0x78 0x7c 0x100 0x104 0x108 0x10c 0x110 0x114 0x118 0x11c] {

                  set z [catch {VHSADCClient__Vhsadac16ReadFpgaReg $ID $offset} rc]
                  if {$z == 0} {
                      puts $f "$ID [format 0x%03x $offset] [format 0x%08x $rc]"
                  } else {
                      set-footer "Error reading FpgaReg offset=[format 0x%03x $offset]: NOT saved"
                  }
              }
          }
      }

      if {$Options(SaveAcqRegs) == 1 && $AcquisitionMethod == 0} {

          foreach ID $boardlist {

             foreach offset [list 0x1000 0x1004 0x1008 0x100c 0x1010 0x1014 0x1018] {

                  set z [catch {VHSADCClient__Vhsadac16ReadFpgaReg $ID $offset} rc]
                  if {$z == 0} {
                      puts $f "$ID [format 0x%03x $offset] [format 0x%08x $rc]"
                  } else {
                      set-footer "Error reading FpgaReg offset=[format 0x%03x $offset]: NOT saved"
                  }
              }
          }
      }

      if {$Options(SaveAcqRegs) == 1 && $UserGUIVersion == 2} {

          foreach ID $boardlist {

            foreach C $chanlist {

              if {$C >= $NumChan($ID)} {break}

              set SaveID $boardID
              set SaveC $chan

              set boardID $ID
              set chan $C

              set z [catch {Read_MWD_Registers} m]

              set boardID $SaveID
              set chan $SaveC

              if {$z == 0} {
                  set reg 0
                  global MWDr$reg 
                  puts $f "$ID MWDr$reg [format 0x%08x [set MWDr$reg]] $C"
                  for {set reg 16} {$reg <= 26} {incr reg} {
                      global MWDr$reg
                      puts $f "$ID MWDr$reg [format 0x%08x [set MWDr$reg]] $C"
                  }
                  foreach reg [list 0x20] {  ;# CFD Threshold (per channel)
                      global Reg$reg
                      puts $f "$ID Reg$reg [format 0x%08x [set Reg$reg]] $C"
                  }
                  foreach reg [list 0x31] {   ;# PreTrigger Position (per card)
                      global Reg$reg
                      if {$C == 0} {puts $f "$ID Reg$reg [format 0x%08x [set Reg$reg]] $C"}
                  }
              } else {
                  set-footer "Error reading MWD parameters: NOT saved"
              }
            }
          }
      }

      close $f

     set-footer "Settings saved to $filename"
}

proc click-restore {} {
    global env boardlist boardID 
    upvar #0 savedirectory  directory
    upvar #0 savefile filename

    	option add *Canvas.background  #c0c020
	set types {
	{ {Data Files} {*.dat} }
	{ {All Files } {*} }
	}

	set f [tk_getOpenFile -parent [self frame-name] \
		  -title "Save File Selection"  \
		  -initialdir $directory -filetypes $types]
	option add *Canvas.background  $env(tkcanvasbackground)
	if {$f == ""} then {return}

      set filename $f   ;#  remember current choice
      set directory [file dirname $f]
      Save_Options

      do-restore
}

proc do-restore {} { 
    global env boardlist boardID Options NumChan
    global AcquisitionMethod UserGUIVersion
    upvar #0 savedirectory  directory
    upvar #0 savefile filename

      set z [catch {open $filename r} f]
      if {$z != 0} then {
          midas-warning "Cannot create/open file $filename\n$f"
          return
      }

      set settings [split [read $f] \n]

      close $f

      set saveIt $boardID

      foreach setting $settings {

          if {$setting == ""} {continue}
          set ID [first $setting]
          set offset [second $setting]
          set vl [third $setting]

          if {[string range $offset 0 3] == "MWDr"} then {

              set boardID $ID

              set reg [string range $offset 4 end]
              if {$reg != 0} {if {$reg < 16 || $reg > 26} {continue}}

              set c [lindex $setting 3]
              if {$c >= $NumChan($ID)} {continue}

              set-footer "Restoring MWD parameter $reg for board $boardID channel $c"
              set z [catch {VHSADCClient__WriteUserReg $c $reg $vl} rc]
              if {$z != 0} {
                  set-footer "Error writing MWD register offset=[format 0x%03x $offset] ($z $rc): NOT restored"
              }
          } else {

          if {[string range $offset 0 2] == "Reg"} then {

              set boardID $ID

              set reg [string range $offset 3 end]

              set c [lindex $setting 3]
              if {$c >= $NumChan($ID)} {continue}

              set-footer "Restoring Register $reg for board $boardID channel $c"
              set z [catch {VHSADCClient__WriteUserReg $c $reg $vl} rc]
              if {$z != 0} {
                  set-footer "Error writing Register offset=[format 0x%03x $offset] ($z $rc): NOT restored"
              }
          } else {

set vl [expr $vl & 0x7fffffff]  ;# I have a problem with the MSB

              insert "setting board $ID register [format "0x%x" $offset] to [format "0x%x" $vl]"
              set z [catch {VHSADCClient__Vhsadac16WriteFpgaReg $ID $offset $vl} rc]
              if {$z != 0} {
                  set-footer "Error writing FpgaReg offset=[format 0x%03x $offset] ($z $rc): NOT restored"
              }
          }

          }
      }

      set boardID $saveIt

      set saveIt $boardID
      foreach boardID $boardlist {
          if {$Options(SaveGains) == 1} {click-gainpgm}
          if {$Options(SaveClocks) == 1} {click-clkpgm}
          if {$AcquisitionMethod == 0} {fpgaSDRAMacq_channelselect}   ;# to setup selected channels
          if {$Options(SaveAcqRegs) == 1 && $UserGUIVersion == 2} {catch {Set_MWD_Registers}}
      }
      set boardID $saveIt

      redisplay

     set-footer "Settings restored from $filename"

}

proc click-clks {f} {                             ;# set all ADC clocks
    global boardlist boardID

    set saveIt $boardID
    foreach ID $boardlist {

       set z [catch {VHSADCClient__Vhsadac16Clock_Frequency $ID $f} rc]
       if {$z != 0} {
            set-footer "Error setting ADC clock frequency; rc = $rc"
       }
    }
    set boardID $saveIt

    update_clock
}

proc click-clk105 {args} {click-clks 105}                             ;# set ADC clocks to 105 MHz

proc click-clk {f} {                             ;# set ADC clock
    global boardID

       set z [catch {VHSADCClient__Vhsadac16Clock_Frequency $boardID $f} rc]
       if {$z != 0} {
            set-footer "Error setting ADC clock frequency; rc = $rc"
       }
}

proc Restore {} {                         ;# used by SETUP
    upvar #0 savefile filename

    if {$filename == "undefined"} {click-restore} else {do-restore}
}

proc click-sdrambasec {v} {
   global  sdrambase sdrambasechoice

   set sdrambase  [last [lindex $sdrambasechoice $v]]
   :modify SDRAMBASET :value [format "0x%05x" $sdrambase]
}

proc click-sdrambaset {v} {
   global  sdrambase
   if {[scan-int $v] == $v && $v >= 0} then {set sdrambase $v}
   :modify SDRAMBASET :value [format "0x%05x" $sdrambase]
   :modify SDRAMBASEC :value -1
}

proc click-sdramaddr {v} {
   global sdramaddr
   if {[scan-int $v] == $v && $v >= 0 && $v < 0x10000} then {set sdramaddr $v}
   :modify SDRAMADDR :value [format "0x%04x" $sdramaddr]
}

proc click-sdramaddrz {args} {click-sdramaddr 0}

proc click-sdramwords {v} {
   global sdramwords
   if {[scan-int $v] == $v && $v >= 0 && $v <= 0x400} then {set sdramwords $v}
   :modify SDRAM WORDS :value [format "0x%02x" $sdramwords]
}

proc  do_sdramdisplay {} {
   global boardID
   global  sdrambase sdramaddr sdramwords
   global DataRAM

   set-footer [self frame-name].sdram "Reading data from SDRAM"

   set baseaddress [expr $sdrambase + $sdramaddr]
   set length [expr $sdramwords*4]
   set DataRAM [VHSADCClient__Vhsadac16Sdram_Read $boardID $baseaddress $length]

   set info ""
   set i 0
   set offset 0
   foreach word $DataRAM {
      if {$i == 0} {append info "[format "0x%08x" [expr $baseaddress + $offset]]: "}
      append info "[format 0x%08x $word]  "
      incr i
      if {$i == 8} then {append info "\n"; set i 0}
      incr offset 4
   }
   :modify sdramINFO :strings $info\n

   clear-footer [self frame-name].sdram

   click-sdramaddr [expr $sdramaddr + ($sdramwords * 4)]
}

proc click-fpgaclk {v i} {
    global ADCClockFreq

   if {$v == 0} {        ;# external clock
       set ADCClockFreq "External"
   } else {
       set ADCClockFreq [string range $i 0 [expr [string length $i] - 4]]
   }

   Save_Options

}

proc click-fpgaframesize {i} {
    global FrameSize

   set FrameSize $i

   Save_Options
}

proc click-fpgasetframesize {} {
    global FrameSize
    global boardID boardlist

    set saveIt $boardID
    foreach ID $boardlist {

        set reg 0x30
        set chan 0
       set z [catch {VHSADCClient__WriteUserReg $chan $reg [expr $FrameSize/2]} rc]       ;# in units of 32 bit data words
       if {$z != 0} {
            set-footer "Error setting frame size; rc = $rc"
       }
    }
    set boardID $saveIt
}

proc do_fpgasetnumchan {} {
    global boardID boardlist chanlist
    global NumChan

    set saveID $boardID
    foreach ID $boardlist {
        set NumChan($ID) 0

        set reg 0x82
        set chan 0
        set rc [VHSADCClient__ReadUserReg $chan $reg]
        set rc [expr $rc & 0xffff]
        if {$rc < 1 || $rc > 16} {set rc 1}    ;# set a default if value return is invalid

        set NumChan($ID) $rc
        set-footer "[self frame-name].fpga" "board $ID detected to support $NumChan($ID) channels"
    }
    set boardID $saveID
}


proc click-fpgasetclk {args} {
    global ADCClockFreq FPGAProgramOption

    if {$ADCClockFreq == "External"} {return}

    if {$FPGAProgramOption == 0} then {click-clk $ADCClockFreq} else {click-clks $ADCClockFreq}
}

proc click-fpgareset {args} {
    global boardID

    set ctrl [VHSADCClient__Vhsadac16ReadFpgaReg $boardID 0x00]  ;#  board control register
    set ctrl [expr $ctrl & 0x000005f8]    ;# remove reserved bits
    VHSADCClient__Vhsadac16WriteFpgaReg $boardID 0x00 [expr $ctrl | 0x100] ;# set INT RST bit  (8)

    after 1000 update_
}

proc click-fpgafile {args} {
    upvar #0 fpgadirectory  directory
    upvar #0 fpgaffile filename

    set filename [first $args]
    set directory [file dirname $filename]
    Save_Options
}

proc click-fpgabrowse {args} {
    global env
    upvar #0 fpgadirectory  directory
    upvar #0 fpgaffile filename

    	option add *Canvas.background  #c0c020
	set types {
	{ {Bitstream Files} {*.bit} }
	{ {All Files } {*} }
	}

	set f [tk_getOpenFile -parent [self frame-name] \
		  -title "FPGA bitstream File Selection"  \
		  -initialdir $directory -filetypes $types]
	option add *Canvas.background  $env(tkcanvasbackground)
	if {$f == ""} then {return}

      set filename $f   ;#  remember current choice
      set directory [file dirname $f]
      Save_Options

      :modify FPGAFILE :value $filename
}

proc do_Copyfpga {file} {
   global env tcl_platform

   if {![info exists env(MIDASTempDir)] || ![file exists $env(MIDASTempDir)]} then {
       switch $tcl_platform(platform) {
          windows {set env(MIDASTempDir) [file join c:/ temp]}
          default {set env(MIDASTempDir) \tmp}
       }
   }

   catch {file mkdir $env(MIDASTempDir}
   catch {file mkdir [file join $env(MIDASTempDir) VHSADCClient]}
   catch {file mkdir [file join $env(MIDASTempDir) VHSADCServer]}

   set destn [file join $env(MIDASTempDir) VHSADCClient [file tail $file]]
   set z [catch {file copy -force $file $destn} f]
   if {$z != 0} then {midas-warning "Cannot copy file $file to $destn\n$f"; return -1}

   set destn [file join $env(MIDASTempDir) VHSADCServer [file tail $file]]
   set z [catch {file copy -force $file $destn} f]
   if {$z != 0} then {midas-warning "Cannot copy file $file to $destn\n$f"; return -1}

   return 0
}

proc do_UpLoadfpga {} {
    upvar #0 fpgaffile file
    global WebServer

    if {$WebServer == "localhost"} {
        set rc [do_Copyfpga $file]
        if {$rc == 0} {set-footer "[self frame-name].fpga"   "Uploaded file $file successful"}
    } else {
        set-footer "[self frame-name].fpga"  "Uploading FPGA bitstream file.... This takes about 10 secs"
        set z [catch {VHSADCClient__VHSADCUpLoad $file} rc]
        clear-footer "[self frame-name].fpga" 

        if {$z == 0} {
            set-footer "[self frame-name].fpga"   "Uploaded file $file successful"
        } else {
            bgerror "Uploading file $file failed\n$rc"
        }
    }
}

proc do_Programfpga {} {
    global boardID boardlist
    global FPGAProgramOption ADCClockFreq
    upvar #0 fpgaffile filename

    foreach ID $boardlist {

        if {$ID != $boardID && $FPGAProgramOption == 0} {continue}

        set-footer "[self frame-name].fpga"  "loading FPGA for board $ID.... This takes about 10 secs"
        set z [catch {VHSADCClient__Vhsadac16FpgaLoad $ID $filename} rc]
        clear-footer "[self frame-name].fpga" 

        if {$z == 0} {
            if {[string range $rc 0 2] == "0x0"} then {
                set-footer "[self frame-name].fpga"   "FpgaLoad for board $ID file $filename successful"
            } else {
                bgerror "FpgaLoad for board $ID file $filename failed\n$rc"
            }
        } else {
             bgerror "FpgaLoad for board $ID file $filename failed\n$rc"
        }
    }

    click-fpgasetclk
    click-fpgasetframesize
    do_fpgasetnumchan
}

proc do_UpLoadProgramfpga {}  {do_UpLoadfpga; do_Programfpga}

proc do_fpgaOption {args} {
    global FPGAProgramOption

    set FPGAProgramOption [lindex $args 1]
}

proc click-clkfile {args} {
    upvar #0 clkdirectory  directory
    upvar #0 clkfile filename

    set filename [first $args]
    set directory [file dirname $filename]
    Save_Options
}

proc click-clkbrowse {args} {
    global env
    upvar #0 clkdirectory  directory
    upvar #0 clkfile filename

    	option add *Canvas.background  #c0c020
	set types {
	{ {Bitstream Files} {*.dat} }
	{ {All Files } {*} }
	}

	set f [tk_getOpenFile -parent [self frame-name] \
		  -title "FS6377 data File Selection"  \
		  -initialdir $directory -filetypes $types]
	option add *Canvas.background  $env(tkcanvasbackground)
	if {$f == ""} then {return}

      set filename $f   ;#  remember current choice
      set directory [file dirname $f]
      Save_Options

      :modify CLKFILE :value $filename
}

proc click-clkload {args} {
    global boardID
    upvar #0 clkdirectory  directory
    upvar #0 clkfile filename

      set z [catch {open $filename r} f]
      if {$z != 0} then {
          midas-warning "Cannot open file $filename\n$f"
          return
      }

      set clkdat [split [read $f] \n]

      close $f

      for {set i 0} {$i < [expr [llength $clkdat]-2]} {incr i 4} {
          set dat [expr ([lindex $clkdat [expr $i+3]] << 24) + \
                        ([lindex $clkdat [expr $i+2]] << 16) + \
                        ([lindex $clkdat [expr $i+1]] << 8) + [lindex $clkdat $i]]
          set rc [VHSADCClient__Vhsadac16WriteFpgaReg $boardID [expr 0x0c + $i] $dat]
      }

      update_clock
}

proc click-boardregisters {args}  {
    global boardID

    insert "VHS-ADC control registers"

    set info ""
    set i 0
    foreach offset [list 0x0 0x8 0xc 0x10 0x14 0x18 0x1c 0x20 0x24 0x28 0x2c 0x30 \
       0x34 0x38 0x3c 0x40 0x44] {

        set z [catch {VHSADCClient__Vhsadac16ReadFpgaReg $boardID $offset} rc]
        if {$z == 0} {
            append info "[format 0x%03x $offset]=[format 0x%08x $rc]   "
        } else {
            set-footer "Error reading FpgaReg offset=[format 0x%03x $offset]"
            append info "[format 0x%03x $offset]=**********   "
        }

        incr i
        if {$i == 4} {append info \n; set i 0}
    }
    insert $info

    insert "User-defined control registers"

    set info ""
    set i 0
    foreach offset [list 0x70 0x74 0x78 0x7c 0x100 0x104 0x108 0x10c 0x110 0x114 0x118 0x11c] {

        set z [catch {VHSADCClient__Vhsadac16ReadFpgaReg $boardID $offset} rc]
        if {$z == 0} {
            append info "[format 0x%03x $offset]=[format 0x%08x $rc]   "
        } else {
            set-footer "Error reading FpgaReg offset=[format 0x%03x $offset]"
            append info "[format 0x%03x $offset]=**********   "
        }

        incr i
        if {$i == 4} {append info \n; set i 0}
    }
    insert $info
}

proc click-fpgaregisters {args}  {
    global boardID

    insert "FPGA SDRAM Acquisition Registers"

    set info ""
    set i 0
    foreach offset [list 0x0 0x4 0x8 0xc 0x10 0x14 0x18 0x1c] {

        set z [catch {VHSADCClient__Vhsadac16ReadFpgaReg $boardID [expr 0x1000 + $offset]} rc]
        if {$z == 0} {
            append info "[format 0x%03x $offset]=[format 0x%08x $rc]   "
        } else {
            set-footer "Error reading FpgaReg offset=[format 0x%03x $offset]"
            append info "[format 0x%03x $offset]=**********   "
        }

        incr i
        if {$i == 4} {append info \n; set i 0}
    }
    insert $info
}

proc click-rtdexregisters {args}  {
    global boardID

    insert "FPGA RTDEX Acquisition Registers"

    set info ""
    set i 0
    foreach offset [list 0x100 0x104 0x108 0x10c] {

        set z [catch {VHSADCClient__Vhsadac16ReadFpgaReg $boardID [expr 0x1000 + $offset]} rc]
        if {$z == 0} {
            append info "[format 0x%03x $offset]=[format 0x%08x $rc]   "
        } else {
            set-footer "Error reading FpgaReg offset=[format 0x%03x $offset]"
            append info "[format 0x%03x $offset]=**********   "
        }

        incr i
        if {$i == 4} {append info \n; set i 0}
    }
    insert $info
}

proc ResetDCM {args} {
    global boardID

    set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID 0x08]  ;# clocks control register
    VHSADCClient__Vhsadac16WriteFpgaReg $boardID 0x08 [expr ($rc & 0x0000ff7f) | 0x80]  ;# set Reset DCM bit

    set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID 0x08]  ;# clocks control register
    if {[expr ($rc >> 7) & 1] == 1} {
        midas-warning "FPGA DCM reset bit NOT cleared"
        VHSADCClient__Vhsadac16WriteFpgaReg $boardID 0x08 [expr $rc & 0x0000ff7f]  ;# unset Reset DCM bit
    }

    update_clock
}

proc click-cdpclk {vl} {
    global boardID

    set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID 0x08]  ;# clocks control register
    VHSADCClient__Vhsadac16WriteFpgaReg $boardID 0x08 [expr ($rc & 0x1f) | ($vl << 5)]  ;# set CD PCLK bit

    update_clock
}

proc click-oepclk {vl} {
    global boardID

    set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID 0x08]  ;# clocks control register
    VHSADCClient__Vhsadac16WriteFpgaReg $boardID 0x08 [expr ($rc & 0x2f) | ($vl << 4)]  ;# set OEPCLK bit

    update_clock
}

proc click-pdpclk {vl} {
    global boardID

    set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID 0x08]  ;# clocks control register
    VHSADCClient__Vhsadac16WriteFpgaReg $boardID 0x08 [expr ($rc & 0x37) | ($vl << 3)]  ;# set Pdown PCLK bit

    update_clock
}

proc click-quadclk {vl} {
    global boardID

    set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID 0x08]  ;# clocks control register
    VHSADCClient__Vhsadac16WriteFpgaReg $boardID 0x08 [expr ($rc & 0x3b) | ($vl << 2)]  ;# set QUADRATURE bit

    update_clock
}

proc click-fpdpclk {vl} {
    global boardID

    set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID 0x08]  ;# clocks control register
    VHSADCClient__Vhsadac16WriteFpgaReg $boardID 0x08 [expr ($rc & 0x3d) | ($vl << 1)]  ;# set FPDP SEL bit

    update_clock
}

proc click-clksrc {vl} {

    if {[FpgaType] == 0} {click-clksrc2 $vl} else {click-clksrc4 $vl}
}

proc click-clksrc2 {vl} {
    global boardID

    set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID 0x08]  ;# clocks control register
    VHSADCClient__Vhsadac16WriteFpgaReg $boardID 0x08 [expr ($rc & 0x3e) | $vl]  ;# set CLK SRC bit

    update_clock
}

proc click-clksrc4 {vl} {
    global boardID

    set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID 0x08]  ;# clocks control register
    VHSADCClient__Vhsadac16WriteFpgaReg $boardID 0x08 [expr ($rc & 0x00000ff8) + 0x2000 + $vl]  ;# set CLK SRC bits

    update_clock
}

proc click-trig {vl} {
    global boardID

    VHSADCClient__Vhsadac16WriteFpgaReg $boardID 0x1c $vl ;# acq/trig control register
}

proc update_gain {} {
    global boardID

    set ctrl [VHSADCClient__Vhsadac16ReadFpgaReg $boardID 0x24]  ;# gain configuration register

    :modify GAINCTRL :value [expr ($ctrl >> 18) & 1]
    :modify GAINALL  :value [expr ($ctrl >> 16) & 1]

    set j 1
    for {set i 0} {$i < 8} {incr i} {
        set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID [expr 0x28 + ($i*4)]]  ;# gain register
        set gain [expr $rc & 0x0fff]
        set en [expr ($ctrl >> ($j-1)) & 1]
        :modify GAINVL$j :value $gain
        :modify GAINEN$j :value $en
        incr j
        set gain [expr ($rc >> 16) & 0xfff]
        set en [expr ($ctrl >> ($j-1)) & 1]
        :modify GAINVL$j :value $gain
        :modify GAINEN$j :value $en
        incr j
    }

    if {[expr ($ctrl >> 17) & 1] != 0} {insert "Warning: Program Gain bit is set"}

}

proc click-gainpgm {args} {
    global boardID

    set ctrl [VHSADCClient__Vhsadac16ReadFpgaReg $boardID 0x24]  ;# gain configuration register
    set ctrl [expr $ctrl & 0x0005ffff]    ;# remove reserved bits
    VHSADCClient__Vhsadac16WriteFpgaReg $boardID 0x24 [expr $ctrl | 0x20000] ;# set PrG lvl bit  (17)

    set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID 0x24]  ;# gain configuration  register
    if {[expr ($rc >> 17) & 1] == 1} {midas-warning "Gain programming bit NOT cleared"}

    update_gain
}

proc click-gainctrl {vl} {
    global boardID

    set ctrl [VHSADCClient__Vhsadac16ReadFpgaReg $boardID 0x24]  ;# gain configuration register
    set ctrl [expr $ctrl & 0x0005ffff]    ;# remove reserved bits
    VHSADCClient__Vhsadac16WriteFpgaReg $boardID 0x24 [expr $ctrl | ($vl << 18)] ;# set PrG Ctrl bit  (18)

    update_gain
}

proc click-gainall {vl} {
    global boardID

    set ctrl [VHSADCClient__Vhsadac16ReadFpgaReg $boardID 0x24]  ;# gain configuration register
    set ctrl [expr $ctrl & 0x0005ffff]    ;# remove reserved bits
    VHSADCClient__Vhsadac16WriteFpgaReg $boardID 0x24 [expr $ctrl | ($vl << 16)] ;# set Ch1 bit  (16)

    update_gain
}

proc click-gainen {ch vl} {
    global boardID

    set ctrl [VHSADCClient__Vhsadac16ReadFpgaReg $boardID 0x24]  ;# gain configuration register
    set ctrl [expr $ctrl & 0x0005ffff]    ;# remove reserved bits
    set new [format 0x%x [expr ($ctrl & ~(1 << ($ch-1))) | $vl << ($ch-1)]] 
    VHSADCClient__Vhsadac16WriteFpgaReg $boardID 0x24 $new       ;# set Ch enable bit

    update_gain
}

proc click-gainvl {ch vl} {
    global boardID

    set i [expr ($ch-1)/2]
    set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID [expr 0x28 + ($i*4)]]  ;# gain register

    if {[expr $ch & 1] == 1} {
        set new [format 0x%x [expr ($rc & 0x0fff0000) + $vl]]
    } else {
        set new [format 0x%x [expr ($rc & 0x0000ffff) + ($vl << 16)]]
    }

    VHSADCClient__Vhsadac16WriteFpgaReg $boardID [expr 0x28 + ($i*4)]] $new    ;# update gain register

    update_gain
}

proc click-clkvl {ch vl} {
    global boardID

   set i [expr (($ch-1)/4)*4]
   set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID [expr 0x0c + $i]] 

   switch [expr $ch-1-((($ch-1)/4)*4)] {
       0  {set vl [expr ($rc & 0xffffff00) | $vl]}
       1  {set vl [expr ($rc & 0xffff00ff) | ($vl << 8)]}
       2  {set vl [expr ($rc & 0xff00ffff) | ($vl << 16)]}
       3  {set vl [expr ($rc & 0x00ffffff) | ($vl << 24)]}
   }

   set rc [VHSADCClient__Vhsadac16WriteFpgaReg $boardID [expr 0x0c + $i] $vl]

   update_clock
}

proc click-clkpgm {args} {
    global boardID

    set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID 0x0c]  ;# programmable clock value registers 1-4
    if {$rc == 0} {return}   ;#  no values set

    set-footer "Programming PLL CLK"

    set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID 0x08]  ;# clocks control register
    VHSADCClient__Vhsadac16WriteFpgaReg $boardID 0x08 [expr ($rc & 0x39) | 0x40]  ;# set PCLK bit

    set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID 0x08]  ;# clocks control register
    if {[expr ($rc >> 6) & 1] == 1} {midas-warning "PLL CLK programming bit NOT cleared"}

    ResetDCM    ;#  also calls update_clock

    clear-footer
}


proc update_clock {} {

    if {[FpgaType] == 0} {update_clock2} else {update_clock4}
}

proc update_clock2 {} {
    global boardID

    set z [catch {:query_Mapped [self frame-name].clock} rc]
    if {$z != 0 || $rc != 1} {return}

    set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID 0x08]  ;# clocks control register

    :modify CLKSRC :value [expr $rc & 1]  ;# CLK SRC

    :modify CDPCLK :value [expr ($rc >> 5) & 1]
    :modify OEPCLK :value [expr ($rc >> 4) & 1]
    :modify PDPCLK :value [expr ($rc >> 3) & 1]
    :modify QUADCLK :value [expr ($rc >> 2) & 1]   ;# QUAD
    :modify FPDPCLK :value [expr ($rc >> 1) & 1]   ;# FPDP SEL

    if {[expr $rc & 0x100] == 0} {:modify DCM :value "NOT locked"} else {:modify DCM :value "Locked"}

    if {[expr $rc & 0x80] != 0} {insert "Warning: Reset FPGA DCM bit is set"}
    if {[expr $rc & 0x40] != 0} {insert "Warning: Program PLL CLK bit is set"}

    set j 1
    for {set i 0} {$i < 4} {incr i} {
        set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID [expr 0x0c + ($i*4)]]  ;# user register
        for {set k 0} {$k < 4} {incr k} {
            :modify CLKVL$j :value [format 0x%02x [expr ($rc >> (($j-1)*8)) & 255]]
            incr j
        }
    }
}

proc update_clock4 {} {
    global boardID

    set z [catch {:query_Mapped [self frame-name].clock} rc]
    if {$z != 0 || $rc != 1} {return}

    set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID 0x08]  ;# clocks control register

    :modify CLKSRC :value [expr $rc & 7]  ;# CLK SRC
    :modify CLKDIV2 :value [expr ($rc >> 15) & 1]

    if {[expr $rc & 0x100] == 0} {:modify DCM :value "NOT locked"} else {:modify DCM :value "Locked"}

    if {[expr $rc & 0x80] != 0} {insert "Warning: Reset FPGA DCM bit is set"}
}

proc update_adc {} {
    global boardID

    set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID 0x20] 

    for {set i 1} {$i <= 16} {incr i} {
       :modify ADC$i :value [expr ($rc >> ($i-1)) & 1]
    }
}

proc update_fpgaacq {} {
    global AcquisitionMethod

   if {$AcquisitionMethod == 0} {update_SDRAMfpgaacq; return}
   if {$AcquisitionMethod == 1} {update_RTDEXfpgaacq; return}

}

proc update_SDRAMfpgaacq {} {
    global boardID

    set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID [expr 0x1000 + 0x00]]

    set StorageFormat [expr ($rc >> 24) & 255]
    if {$StorageFormat == 1} {:modify FPGAACQSF :value 0}
    if {$StorageFormat == 2} {:modify FPGAACQSF :value 1}
    if {$StorageFormat == 4} {:modify FPGAACQSF :value 2}

    set SourceID [expr ($rc >> 8) & 255]
    :modify FPGAACQSRC :value -1
    if {$SourceID == 16} {:modify FPGAACQSRC :value 0}
    if {$SourceID == 6} {:modify FPGAACQSRC :value 1}
    if {$SourceID == 32} {:modify FPGAACQSRC :value 2}

    set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID [expr 0x1000 + 0x10]]

    :modify FPGAACQTRIGSEL :value [expr $rc & 255]
    :modify FPGAACQTRIGSRC :value [expr ($rc >> 19) & 3]
    :modify FPGAACQTRIGPOL :value [expr ($rc >> 16) & 1]
    :modify FPGAACQTRIGRUN :value [expr ($rc >> 17) & 1]
    :modify FPGAACQTRIGMDE :value [expr ($rc >> 18) & 1]

    :modify FPGAACQTRIGSTATE :value [expr ($rc >> 24) & 15]

    set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID [expr 0x1000 + 0x14]]
    :modify FPGAACQTRIGLVL :value $rc

    set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID [expr 0x1000 + 0x18]]
    :modify FPGAACQTRIGDLY :value $rc

    set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID [expr 0x1000 + 0x08]]
    :modify FPGAACQFRAMESIZE :value $rc

    set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID [expr 0x1000 + 0x0c]]
    :modify FPGAACQFRAMECNT :value [expr $rc]

    set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID [expr 0x1000 + 0x04]]
    for {set i 1} {$i <= 16} {incr i} {
          :modify FPGAACQCH$i :value [expr ($rc >> ($i-1)) & 1]
    }

    :modify FPGAACQFERR :value [expr ($rc >> 31) & 1]
    :modify FPGAACQACT :value [expr ($rc >> 30) & 1]
    :modify FPGAACQSAMPLES :value [expr ($rc >> 16) & 31]

}

proc click-rtdexacq {offset vl i} {
    global boardID WriteIF

      set chan 0

        set regvl [expr ($chan << 28) + ($offset << 16) + $vl]        ;#    build data for register
#insert-debug "Writing [format 0x%08x $regvl] to offset $WriteIF"
        VHSADCClient__Vhsadac16WriteFpgaReg $boardID $WriteIF $regvl        ;#    write to user register

}

proc update_RTDEXfpgaacq {} {
    global boardID ReadIF

   if {![:query_Mapped [self frame-name].fpgaacq]} {return}

    set-busy

      set chan 0

      for {set offset 32} {$offset < 42} {incr offset} {

        set regvl [expr ($chan << 28) + ($offset << 16)]              ;#    build data for register
#insert-debug "Writing [format 0x%08x $regvl] to offset $ReadIF"
        VHSADCClient__Vhsadac16WriteFpgaReg $boardID $ReadIF $regvl ;#    write to user register

        set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID $ReadIF]  ;#  read user register
#insert-debug "Read of offset $ReadIF returned [format 0x%08x $rc]"
        :modify RTDEXACQ$offset :value [format 0x%x [expr $rc & 0x0000ffff]]
      }

    clear-busy
}

proc fpga_auto_update {args} {
    global boardID

    if {![:query_Mapped [self frame-name].fpgaacq]} {return}

    set-busy

    set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID [expr 0x1000 + 0x10]]

    :modify FPGAACQTRIGRUN :value [expr ($rc >> 17) & 1]
    :modify FPGAACQTRIGSTATE :value [expr ($rc >> 24) & 15]
    
    clear-busy

   after 1000 fpga_auto_update
}
    

proc click-fpgaacq {w vl i} {
    global boardID

    switch -glob $w {

       SF  {
           set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID [expr 0x1000 + 0x00]]
           if {$i == "Single"} {set opt 1}
           if {$i == "Double"} {set opt 2}
           if {$i == "Quad"} {set opt 4}
           VHSADCClient__Vhsadac16WriteFpgaReg $boardID [expr 0x1000 + 0x00] \
                [expr ($rc & 0x00ffff00) | ($opt << 24)]
       }

       SRC  {
           set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID [expr 0x1000 + 0x00]]
           if {$i == "Internal"} {set opt 16}
           if {$i == "VHS-ADC"} {set opt 6}
           if {$i == "FPGALink"} {set opt 32}
           VHSADCClient__Vhsadac16WriteFpgaReg $boardID [expr 0x1000 + 0x00] \
                [expr ($rc & 0xffff0000) | ($opt << 8)]
       }

       FRAMESIZE  {
           VHSADCClient__Vhsadac16WriteFpgaReg $boardID [expr 0x1000 + 0x08] $vl
       }

       TRIGLVL  {
           VHSADCClient__Vhsadac16WriteFpgaReg $boardID [expr 0x1000 + 0x14] $vl
       }

       TRIGDLY  {
           VHSADCClient__Vhsadac16WriteFpgaReg $boardID [expr 0x1000 + 0x18] $vl
       }

       TRIGSEL  {
           set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID [expr 0x1000 + 0x10]]
           VHSADCClient__Vhsadac16WriteFpgaReg $boardID [expr 0x1000 + 0x10] \
                [expr ($rc & 0x00ff0000) | $vl]
       }

       TRIGPOL  {
           set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID [expr 0x1000 + 0x10]]
           VHSADCClient__Vhsadac16WriteFpgaReg $boardID [expr 0x1000 + 0x10] \
                [expr ($rc & 0x001e00ff) | ($vl << 16)]
       }

       TRIGMDE  {
           set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID [expr 0x1000 + 0x10]]
           VHSADCClient__Vhsadac16WriteFpgaReg $boardID [expr 0x1000 + 0x10] \
                [expr ($rc & 0x001b00ff) | ($vl << 18)]
       }

       TRIGSRC  {
           set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID [expr 0x1000 + 0x10]]
           VHSADCClient__Vhsadac16WriteFpgaReg $boardID [expr 0x1000 + 0x10] \
                [expr ($rc & 0x000700ff) | ($vl << 19)]
       }

       CH*  {
           set i [expr [string range $w 2 end] - 1]
           set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID [expr 0x1000 + 0x04]]
           set mask [expr ~(1 << $i)]
           VHSADCClient__Vhsadac16WriteFpgaReg $boardID [expr 0x1000 + 0x04] \
                [expr ($rc & 0x0000ffff) & $mask | ($vl << $i)]
           fpgaSDRAMacq_channelselect
       }

       ALL  {
           VHSADCClient__Vhsadac16WriteFpgaReg $boardID [expr 0x1000 + 0x04] 0x0000ffff
           fpgaSDRAMacq_channelselect
       }

       NONE  {
           VHSADCClient__Vhsadac16WriteFpgaReg $boardID [expr 0x1000 + 0x04] 0x00000000
           fpgaSDRAMacq_channelselect
       }

    }

    update_fpgaacq
}

proc fpgaSDRAMacq_channelselect {} {
    global boardID

    set register [expr 0x1000 + 0x04]   ;# data status|data control|data select register

    set-busy

    set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID $register]
    VHSADCClient__Vhsadac16WriteFpgaReg $boardID $register 0x00800000    ;# clear storage indexes

    for {set i 0} {$i < 16} {incr i} {
       if {[expr ($rc >> $i) & 1] == 1} {
           set vl [expr ($rc & 0x0000ffff) | ($i << 16)]
           VHSADCClient__Vhsadac16WriteFpgaReg $boardID $register $vl
       }
    }

    clear-busy
}

proc update_ {} {
    global boardID

    set rc [VHSADCClient__Vhsadac16ReadFpgaReg $boardID 0x1c]   ;# acquisition and trigger control

    :modify ACQ  :value [expr $rc & 1]
    :modify TRIG :value [expr ($rc >> 3) & 1]
}

proc click-boardinfo {args} {
    global boardID

    :modify LOG :empty T      ;#  empty LOG window  

#    Get FPGA Revision

    set rc1 [expr [VHSADCClient__Vhsadac16ReadFpgaReg $boardID 0] >> 16 & 255]
#    insert "VHSADCGetFPGARevision returned $rc1"

#    Get FPGA Board Status and Control Information

    set rc2 [VHSADCClient__Vhsadac16ReadFpgaReg $boardID 0]
#    insert "Vhsadac16_ReadFpgaReg VHSVIRTEX_STATUS_CTRL returned $rc2"

#    Get CPLD Registers Information

    set rc3 [VHSADCClient__Vhsadac16GetCpldRegs $boardID]
#    insert "Vhsadac16_Vhsadac16GetCpldRegs returned $rc3"

    set virtex_pgm [lindex $rc3 0]
    set adc_speed_status [lindex $rc3 1]
    set cpld_status [lindex $rc3 2]
    set cpld_revision [lindex $rc3 3]
    set pcb_revision [lindex $rc3 4]
    set sm_board_id [lindex $rc3 5]

#    print board info

    switch $sm_board_id {
         0x82  {set infos "VHS-ADC 8 - Virtex II 3000"}
         0x83  {set infos "VHS-ADC 8 - Virtex II 4000"}
         0x84  {set infos "VHS-ADC 8 - Virtex II 6000"}
         0x85  {set infos "VHS-ADC 8 - Virtex II 8000"}
         0x8a  {set infos "VHS-ADC 8 - XC4LX40"}
         0x8b  {set infos "VHS-ADC 8 - XC4LX60"}
         0x8c  {set infos "VHS-ADC 8 - XC4LX80"}
         0x8d  {set infos "VHS-ADC 8 - XC4LX100"}
         0x8e  {set infos "VHS-ADC 8 - XC4LX160"}
         0x8f  {set infos "VHS-ADC 8 - XC4SX55"}
         default  {set infos "Unknown"}
    }

    insert "Slot\[$boardID\]: Main Board:\t\t\t $infos"

    set uifpga_daughter_present [expr ($rc2 & 0x00000200) >> 9]
    set uidaughter_id [expr ($rc2 & 0x00000004) >> 2]
    set uidaughter_present [expr ($rc2 & 0x00000002) >> 1]

    if {$uifpga_daughter_present == 0} {
        if {$uidaughter_present == 1} {
            if {$uidaughter_id == 0} {    ;# daughter is a DAC
                set infos "VHS-DAC 8"
            } else {
                set infos "VHS-ADC 8"
            }
        } else {
            set infos "No daughter board present"
        }
    } else {
        set infos "FPGA hardware module"
    }

    insert "Slot\[$boardID\]: Hardware module:\t\t $infos"
    insert "Slot\[$boardID\]: FPGA Core Revision:\t\t $rc1"

#    Get SDRAM Info

#    set rc4 [VHSADCClient__Vhsadac16Sdram_AutoInit $boardID]
#    insert "Vhsadac16_Vhsadac16Sdram_AutoInit returned $rc4"

#    set rc4 [VHSADCClient__Vhsadac16Sdram_DualPageMode $boardID]
#    insert "Vhsadac16_Vhsadac16Sdram_DualPageMode returned $rc4"

    set uisdram_frequency [VHSADCClient__Vhsadac16Sdram_Frequency $boardID]
#    insert "Vhsadac16_Vhsadac16Sdram_Frequency returned $uisdram_frequency"

    if {$uisdram_frequency != 0xdeadbeef} {
        set uisdram_size [VHSADCClient__Vhsadac16Sdram_Size $boardID]
#        insert "Vhsadac16_Vhsadac16Sdram_Size returned $uisdram_size"
    } else {
        set uisdram_frequency =1
    }

#    print SDRAM frequency and size

    if {$uisdram_frequency != -1}  {    ;#  if SDRAM is initialised
        insert "Slot\[$boardID\]: SDRAM Frequency:\t\t [expr $uisdram_frequency/1000000.0] MHz"
        insert "Slot\[$boardID\]: SDRAM Size:\t\t\t [expr (4.0 * $uisdram_size)/(1024*1024)] MB"
    } else {
        insert "Slot\[$boardID\]: SDRAM Not initialised"
    }

#    print ADC max speed

    switch [expr $adc_speed_status & 3] {
         0  {set infos "65 MHz"}
         1  {set infos "80 MHz"}
         2  {set infos "105 MHz"}
         default  {set infos "Unknown"}
    }

    insert "Slot\[$boardID\]: ADC Maximum Speed:\t\t $infos"

#    print ADC analog bandwidth

    switch [expr ($cpld_status & 0x18) >> 3] {
         0  {set infos "A - BW = 30MHz,DR = 26dB"}
         1  {set infos "B - BW = 100MHz,DR = 17.5dB"}
         2  {set infos "C - BW = 200MHz,DR = 14dB"}
         default  {set infos "Unknown"}
    }

    insert "Slot\[$boardID\]: Programmable Gain Option:\t $infos"

#    print CPLD version

    insert "Slot\[$boardID\]: CPLD Revision:\t\t $cpld_revision"

#    print PCB revision

    insert "Slot\[$boardID\]: PCB Hardware Revision:\t $pcb_revision"

}

#    The Options are experiment specific and normally used by SYS
#    However the same code as for the session preferences is used which allows for wider use in the future.

proc get_Option_file {} {
    global env

#    look for the Option file
    set pd $env(EXPERIMENTBASE)
    if {![file exists $pd]} then {
         insert-log "$pd does not exist - Join an experiment!"
         return {}
    }
    if {![file isdirectory $pd]} then {
         insert-log "$pwd is not useable - it is not a directory"
         return {}
    }
    set pf [file join $pd MidasOptions]
    set z [catch {set f [open $pf r]} m]
    if {$z != 0} then {
        set z [catch {set f [open $pf w]} m]
        if {$z != 0} then {
            insert-log "Unable to access $pf - $m"
            return 1
        }
        close $f
        return {}
    }

#    and read it
    set lines [split [read $f] "\n"]
    close $f

    return $lines
}

proc put_Option_file {lines} {
   global env

#    look for the Option file
    set pd $env(EXPERIMENTBASE)
    if {![file exists $pd]} then {
         insert-log "$pd does not exist - Join an experiment!"
         return {}
    }
    if {![file isdirectory $pd]} then {
         insert-log "$pwd is not useable - it is not a directory"
         return {}
    }
    set pf [file join $pd MidasOptions]
    set z [catch {set f [open $pf w]} m]
    if {$z != 0} then {
        insert-log "Unable to access $pf - $m"
        return 1
    }

    foreach item $lines {
      if {$item != ""} {puts $f $item}
    }

    close $f
    return 0
}

#    save all Options for the current task
proc update_Options {} {
   global env Options

   set lines [get_Option_file]
   set key [Get_My_Name]
   set range [find-keyword $lines $key]
   if {[llength $range] > 0} then {
       set lines [lreplace $lines [first $range] [expr [second $range] + 1]]
   }

   lappend lines "## $key"
   foreach item [array names Options] {
       lappend lines "$item $Options($item)"
   }
   lappend lines "##"

   set z [put_Option_file $lines]
   return $z
}

#    load all Options for the current task
proc load_Options {} {
   global env Options

   set lines [get_Option_file]
   set key [Get_My_Name]
   set range [find-keyword $lines $key]

   if {[llength $range] > 1} then {
       set stuff [lrange $lines [expr [first $range]+1] [second $range]]
       foreach item $stuff {
           set Options([head $item]) [tail $item]
       }
   }
}

#    return task Options
proc get_Options {task} {
   global env

   set lines [get_Option_file]
   set key $task
   set range [find-keyword $lines $key]

   set Options {}

   if {[llength $range] > 1} then {
       set stuff [lrange $lines [expr [first $range]+1] [second $range]]
       foreach item $stuff {
           lappend Options [list [first $item] [second $item]]
       }
   } else {
       lappend Options [list Options.filename  "[file join $env(HOME) MidasOptions Options]"]
   }
   return $Options
}

#    remaining procedures are to display the waveform written (.bin file) by the Lyrtech utility

proc update_extras {args} {}

proc click-waveform {args} {
    global waveform directory

    set  waveform [first $args]
    set directory [file dirname $file]
}

proc click-browsewaveform {args} {
    global waveform
    global env  directory

    	option add *Canvas.background  #c0c020
	set types {
	{ {Waveform Files} {*.bin} }
	{ {All Files } {*} }
	}

	set file [tk_getOpenFile -parent [self frame-name] \
		  -title "Waveform File Selection"  \
		  -initialdir $directory -filetypes $types]
	option add *Canvas.background  $env(tkcanvasbackground)
	if {$file == ""} then {return}

	set waveform $file   ;#  remember current choice
        set directory [file dirname $file]

       :modify WAVEFORM :value $waveform
}

proc click-showwaveform {args} {
    global waveform

    insert "reading $waveform"

    set z [ catch {open $waveform r} f ]
    if { $z != 0 } then {midas-warning "Unable to open file $waveform\n$f"; return}

    fconfigure $f -translation binary -encoding binary

    set data [read $f]

    close $f

    set z [catch {binary scan $data s* r} m]

    midas-check-for-temp

    set file [file join [pwd] VHS.txt]

    insert "creating $file"

    set z [ catch {open $file w} f ]
    if { $z != 0 } then {midas-warning "Unable to open file $file\n$f"; return}

    foreach item $r {puts $f $item}

    close $f

    set destn [file join [pwd] VHS[date +%Y%h%d-%H.%M.%S]]

    insert "converting $file to $destn"

    set z [catch {exec ascii-to-eg -from $file -to $destn -format y} m]

    if { $z != 0 } then {midas-warning "Error converting file $file\n$m"; return}

    insert "showing $destn"

    send sd show $destn

}


loadsource usergui       ;#  contains the User FPGA register GUI
